function [xsol,fsol,exitflag,output] = sounds(fun,x0,algo,algoparams,...
maxevals,ftarget,tolf,seed,varargin)
%
%	A wrapper for calling the sounds (Second-Order UNconstrained optimization 
%	via Direct-Search) algorithms.
%
%	xsol = sounds(fun,x0) applies one of the SOUNDS algorithms to minimize the 
%	function fun starting at x0, and returns the best solution point obtained 
%	xsol.
%
%	xsol = sounds(fun,x0,algo) applies the SOUNDS-type algorithm specified by 
%	the algo parameter, with its default settings.
%
%	xsol = sounds(fun,x0,algo,algoparams) specifies the parameters for the 
%	algorithm  (set algoparams=[] for the default values).
%
%	xsol = sounds(fun,x0,algo,algoparams,maxevals) enforces a limit on the 
%	number of function evaluations used by the method.
%
%	xsol = sounds(fun,x0,algo,algoparams,maxevals,ftarget) specifies a target 
%	value to be reached by the algorithm (set ftarget=[] to ignore this 
%	parameter).
%
%	xsol = sounds(fun,x0,algo,algoparams,maxevals,ftarget,tolf) will stop the
%	method whenever a point x is reached such that 
%		fun(x)-ftarget < tolf*(fun(x)-fun(x0))
%	(set tolf=[] for the default value - only used if ftarget not empty).
%
%	xsol = sounds(fun,x0,algo,algoparams,maxevals,ftarget,tolf,seed) 
%	initializes the random seed generator with the value in seed (set 
%	seed=[] for the default value).
%
%	[xsol,fsol] = sounds(fun,x0,...) returns the best solution xsol and the 
%	corresponding value fsol=fun(xsol).
%
%	[xsol,fsol,exitflag] = sounds(fun,x0,...) additionally returns a flag 
%	related to the outcome of the algorithmic run.
%
%	[xsol,fsol,exitflag,output] = sounds(fun,x0,...) additionally returns a 
%	structure containing detailed information about the run.
%
%	----------------
%	Input parameters
%	----------------
%
%	fun: Objective function handle
%
%	x0: Initial point - Must be provided
%
%	algo: Algorithmic choice - Can be empty, 'sds' by default
%		'sds': Symmetrized Direct Search, a weakly second-order convergent 
%		method
%		'ahds': Approximate Hessian Direct Search, a second-order globally 
%		convergent variant with complexity guarantees
%		'dsds': Decoupled Step Direct Search, a second-order globally 
%		convergent variant with improved complexity properties
%		Default value if not assigned: 'sds'.
%
%
%	algoparams: Various parameters for the algorithm 
%		steps0: Initial value(s) for the step size(s)
%			For sds/ahds, steps0 is expected to be a positive real. 
%			Default value if not assigned: 1
%			For dsds, steps0 is expected to be a vector of length 2 of 
%			positive real values:
%				steps0(1): First-order step size
%				steps0(2): Second-order step size
%			Default value if not assigned: [1 1] 
%		typesets: Type of polling sets used for first and second-order polling.
%			For sds/ahds, typesets must be an integer.
%				Default value if not assigned: 0
%			For dsds, typesets must be an integer vector of length 2.
%				Default value if not assigned: [-1 1]	
%			typesets(1) indicates the type of the first-order polling sets:
%				0: [I -I] (coordinate positive basis) - Default value
%				1: [Q -Q] (fixed rotation of the coordinate positive basis)	
%				2: V_{n+1} (minimal positive basis with uniform angles)
%				3: Q*V_{n+1} (fixed rotation of the minimal positive basis)
%				4: [-e I] (minimal positive basis based on coordinate 
%				directions)
%				Values accepted for sds and dsds only:
%				-1: [dk -dk] (One direction uniformly generated on the unit 
%				sphere at every iteration)
%				-p, where p is an integer bigger than or equal to 2: Dk (set of 
%				p independent directions uniformly distributed on the unit 
%				sphere)
%			typesets(2) indicates the basis used for second-order polling:
%				0: I (identity matrix)
%				1: Random shuffle of the identity matrix
%				2: Q (fixed rotation of the identity matrix)
%		forcings: denote the monomial function(s) used as measure(s) of 
%		sufficient decrease.
%			For sds/ahds, forcing is a nonnegative real.
%				Default value if not assigned: 3.
%			For dsds, forcing is a two-dimensional vector of nonnegative reals.
%				Default value if not assigned: [2 3].
%			In both cases, possible values are:
%				0: Simple decrease (the forcing is zero)
%				q>0: Sufficient decrease of the form rho(t) = c*t^q
%
%	maxevals: Maximum allowed number of function evaluations.
%		Default value if not assigned: 2000*length(x0)
%
%	ftarget: A priori targeted function value
%		Default value if not assigned: -Infty
%
%	tolf: Tolerance for targeted value stopping criterion
%		Default value if not assigned: 1e-3
%
%	seed: Integer used to initialize the random number generator
%		Default value if not assigned: 1
%
%	-----------------
%	Output parameters
%	-----------------
%
%   xsol,fsol: Best point xsol and its value fsol found by the method
%
%   exitflag: flag indicating what happened in the method
%       0: Method stopped because the target function value was
%       reached
%       1: Method stopped because the step size was too small
%       (expected behaviour)
%       2: Method stopped because the budget of function evaluations
%       was exceeded
%       -1: An error occurred and the method did not complete its
%       task
%
%   output:  structure containing the following information
%       funcCount: number of function evaluations performed in total
%       nits: numbers of iterations
%			nits(1) is the total number of iterations
%			nits(2) is the number of successful iterations due to 
%			first-order polling
%			nits(3) is the number of successful iterations due to
%			second-order polling with the basis
%			nits(4) is the number of successful iterations due to
%			crossed polling
%			nits(5) is the number of successful iterations due to
%			the approximate eigenvectors
%		hist: array of size (n+1)*nits(2) containing all the iterates 
%		generated by the method
%			hist(1,:) contains the indexes of successful iterations
%			hist(2:(n+1),:) is the corresponding matrix of iterates
%
%	C. W. Royer - September 27, 2017
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% %
if (nargin<2)
	error('sounds: At least two inputs required');
end
%
if isempty(x0)
	error('sounds: Empty initial point, please input a feasible vector');
else
	x0 = x0(:);
end
%
if (nargin >= 3)
	if isempty(algo)
		algo = 'sds';
	elseif ~ismember(algo,{'sds','ahds','dsds'})
		error('sounds: Unknown algorithmic choice.');
	end
else
	algo = 'sds';
end
%
if (nargin>=4)
	if isempty(algoparams)
		if ismember(algo,{'sds','ahds'})
			algoparams.steps0 = 1;
			algoparams.typesets = 0;
			algoparams.forcings = 3;
		else
			algoparams.steps0 = [1 1];
			algoparams.typesets = [0 0];
			algoparams.forcings = [2 3];
		end
	elseif ~isfield(algoparams,{'steps0','typesets','forcing'})
		error('sounds: Unknown fields for algoparams.');
	else
		if ~isfield(algoparams,'steps0')
			if strcmp(algo,'dsds')
				algoparams.steps0 = [1 1];
			else
				algoparams.steps0 = 1;
			end
		else
			if ~isempty(find(algoparams.steps0 <=0))
				error('sounds: steps0 must contain only positive values.');
			end
		end
%
		if ~isfield(algoparams,'typesets')
			if strcmp(algo,'dsds')
				algoparams.typesets = [0 0];
			else
				algoparams.typesets = 0;
			end
		else
			okayset = (mod(algoparams.typesets(1),1)==0);
			if strcmp(algo,'ahds')
				okayset = okayset && (ismember(algoparams.typesets(1),...
				[0 1 2 3 4]));
			else
				okayset = okayset && ((ismember(algoparams.typesets(1),...
				[0 1 2 3 4])) || (algoparams.typesets(1)<0));
			end
			if ~okayset
				error('sounds: unvalid choice for typeset(1) - see help.');
			end
			if strcmp(algo,'dsds')
				if ~ismember(algoparams.typesets(2),[0 1 2])
					error('sounds: unvalid choice for typesets(2)');
				end	
			end
		end
%
		if ~isfield(algoparams,'forcings')
			if strcmp(algo,'dsds')
				algoparams.forcings = [2 3];
			else
				algoparams.forcings = 3;
			end
		else
			if ~isempty(find(algoparams.forcings < 0))
				error('sounds: Forcing parameter cannot be negative');
			end
		end
	end
else
	if ismember(algo,{'sds','ahds'})
		algoparams.steps0 = 1;
		algoparams.typesets = 0;
		algoparams.forcings = 3;
	else
		algoparams.steps0 = [1 1];
		algoparams.typesets = [0 0];
		algoparams.forcings = [2 3];
	end
end
%
if (nargin >= 5)
	if isempty(maxevals)
		maxevals = 2000*length(x0);
	else
		if ~isinteger(maxevals) || (maxevals < 0)
			error('sounds: maxevals must be a positive integer.');
		elseif (maxevals==0)
			warning('sounds: zero budget for function evaluations');
		end
	end
else
	maxevals = 2000*length(x0);
end
%
if (nargin >= 6)
	if isempty(ftarget)
		ftarget = -Inf;
	end
else
	ftarget = -Inf;
end
%
if (nargin >= 7)
	if isempty(tolf)
		tolf = 1e-3;
	elseif (tolf < 0)
		error('sounds: tolf must be nonnegative.');
	end
else
	tolf = 1e-3;
end
%
if (nargin >= 8)
	if isempty(seed)
		seed = 1;
	elseif (seed<0)
		error('sounds: seed must be nonnegative.');
	end
else
	seed = 1;
end
%
% Calling the sounds method
[xsol,fsol,exitflag,output] = soundsalgo(fun,x0,algo,algoparams,maxevals,...
ftarget,tolf,seed);
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
