function [xsol,fsol,flag,output] = soundsalgo(fun,x0,algo,algoparams,...
maxevals,ftarget,tolf,seed)
%
%	Second-Order UNconstrained optimization via Direct Search algorithms.
%
%	Implementation of direct-search methods exploiting second-order 
%	information for unconstrained optimization problems.
%
%	XSOL = soundsalgo(FUN,X0,ALGO,ALGOPARAMS,MAXEVALS,FTARGET,TOLF,SEED) 
%	returns the approximate solution to the problem
%		min FUN(x),
%	obtained by the chosen direct-search method with second-order aspects.
%
%	----------------
%	Input parameters
%	----------------
%
%   fun: Objective function handle
%
%   x0: Initial point - Must be provided
%
%   algo: Algorithmic choice - Can be empty, 'sds' by default
%       'sds': Symmetrized Direct Search, a weakly second-order convergent 
%       method
%       'ahds': Approximate Hessian Direct Search, a second-order globally 
%       convergent variant with complexity guarantees
%       'dsds': Decoupled Step Direct Search, a second-order globally 
%       convergent variant with improved complexity properties
%       Default value if not assigned: 'sds'.
%
%
%   algoparams: Various parameters for the algorithm 
%       steps0: Initial value(s) for the step size(s)
%           For sds/ahds, steps0 is expected to be a positive real. 
%           Default value if not assigned: 1
%           For dsds, steps0 is expected to be a vector of length 2 of 
%           positive real values:
%               steps0(1): First-order step size
%               steps0(2): Second-order step size
%           Default value if not assigned: [1 1] 
%       typesets: Type of polling sets used for first and second-order polling.
%           For sds/ahds, typesets must be an integer.
%               Default value if not assigned: 0
%           For dsds, typesets must be an integer vector of length 2.
%               Default value if not assigned: [-1 1]
%           typesets(1) indicates the type of the first-order polling sets:
%               0: [I -I] (coordinate positive basis) - Default value
%               1: [Q -Q] (fixed rotation of the coordinate positive basis)
%               2: V_{n+1} (minimal positive basis with uniform angles)
%               3: Q*V_{n+1} (fixed rotation of the minimal positive basis)
%               4: [-e I] (minimal positive basis based on coordinate 
%               directions)
%               Values accepted for sds and dsds only:
%               -1: [dk -dk] (One direction uniformly generated on the unit 
%               sphere at every iteration)
%               -p, where p is an integer bigger than or equal to 2: Dk (set of 
%               p independent directions uniformly distributed on the unit 
%               sphere)
%           typesets(2) indicates the basis used for second-order polling:
%               0: I (identity matrix)
%               1: Random shuffle of the identity matrix
%               2: Q (fixed rotation of the identity matrix)
%       forcings: denote the monomial function(s) used as measure(s) of 
%       sufficient decrease.
%           For sds/ahds, forcing is a nonnegative real.
%               Default value if not assigned: 3.
%           For dsds, forcing is a two-dimensional vector of nonnegative reals.
%               Default value if not assigned: [2 3].
%           In both cases, possible values are:
%               0: Simple decrease (the forcing is zero)
%               q>0: Sufficient decrease of the form rho(t) = c*t^q
%
%   maxevals: Maximum allowed number of function evaluations.
%       Default value if not assigned: 2000*length(x0)
%
%   ftarget: A priori targeted function value
%       Default value if not assigned: -Infty
%
%   tolf: Tolerance for targeted value stopping criterion
%       Default value if not assigned: 1e-3
%
%   seed: Integer used to initialize the random number generator
%       Default value if not assigned: 1
%
%
%	-----------------
%	Output parameters
%   -----------------
%
%   xsol,fsol: Best point xsol and its value fsol found by the method
%
%   flag: flag indicating what happened in the method
%       0: Method stopped because the target function value was
%       reached
%       1: Method stopped because the step size was too small
%       (expected behavior)
%       2: Method stopped because the budget of function evaluations
%       was exceeded
%       -1: An error occurred and the method did not complete its
%       task
%
%   output:  structure containing the following information
%       funcCount: number of function evaluations performed in total
%       nits: numbers of iterations
%           nits(1) is the total number of iterations
%           nits(2) is the number of successful iterations due to 
%           first-order polling
%           nits(3) is the number of successful iterations due to
%           second-order polling with the basis
%           nits(4) is the number of successful iterations due to
%           crossed polling
%           nits(5) is the number of successful iterations due to
%           the approximate eigenvectors
%       hist: array of size (n+1)*nits(2) containing all the iterates 
%       generated by the method
%           hist(1,:) contains the indexes of successful iterations
%           hist(2:(n+1),:) is the corresponding matrix of iterates
%
%	C. W. Royer - September 27, 2017
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
	format compact
	format long
	n = length(x0);
	rand('state',seed);
	flag = -1;
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Algorithmic parameters
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
	decds = strcmp(algo,'dsds');
	gamma = 1.1;
	theta = 0.5;
	alpha0 = algoparams.steps0(1);
	p = algoparams.forcings(1);
	tol_alpha = 1e-10;
	alphamax = 20*alpha0;
	if decds
		beta0 = algoparams.steps0(2);
		gammab = 1.1;
		thetab = 0.5;
		q = algoparams.forcings(2);
		tol_beta = 1e-10;
		betamax = 20*beta0;
	end
%
	typesets=algoparams.typesets;
%
%%%%%%%%%%%%%%%%%%% Preparing for main loop %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
	nits = zeros(1,5);
	nevalf = 0;
	x = x0;
	f = feval(fun,x);
	diff = f-ftarget;
	nevalf = nevalf+1;
	hist = [0 ; x0];
	alpha = alpha0;
	if decds
		beta = beta0;
	end
%
	if ~isinf(ftarget)
		stopping = (f > ftarget+tol*diff);
	else
		stopping = 0;
	end
%
%	Computation of the first polling set if deterministic
	D = [];
	symmetric = 1;
	switch typesets(1)
		case 0
			D = eye(n);
			perm = randperm(n);
			D = D(:,perm);
			D = [D -D];
		case 1
			v = 2*rand(n,1)-1;
			D = pss_oppbasis(v);
		case 2
			v = zeros(n,1);
			D = pss_unifangles(v);
			perm = randperm(n+1);
			D = D(:,perm);
			symmetric = 0;
		case 3
			v = 2*rand(n,1)-1;
			D = pss_unifangles(v);
			symmetric = 0;
		case 4
			e = -ones(n,1)	
			e = e/norm(e);
			D = [eye(n) e];
			perm = randperm(n);
			D = D(:,perm);
			symmetric = 0;
		case -1
			d = 2*rand(n,1)-1;
			d = d/norm(d);
			D = [d -d];
		otherwise
			for i=1:abs(typesets(1))
				d = 2*rand(n,1)-1;
				D = [D d];
			end
			symmetric = 0;
	end
%
	if ~isempty(D)
		cardD = size(D,2);
		icycl1 = 1;
	end
%
%	Computation of the second-order polling set (deterministic, a basis) if the 
%	algorithm is not sds or ahds (in the latter case, the basis is extracted 
%	from the polling set)
	if decds
		switch typesets(2)
			case 0
				B = eye(n);
			case 1
				B = eye(n);
				perm = randperm(n);
				B = B(:,perm);
			case 2
				v = 2*rand(n,1)-1;
				[B,~] = qr(v);
		end
%	
	icycl2 = 1;
	end
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Main loop
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
	while ~stopping
%
		nits(1) = nits(1) + 1;
%
%		Forcing function
%
		if (p==0)
			rho1 = 0;
		else
			rho1 = 1e-3*(alpha)^p;
		end
%
		if decds
			if (q==0)
				rho2 = 0;
			else
				rho2 = 1e-3*(beta)^q;
			end
		end
%
%		Auxiliary variables
%
		pbinf = 0;% Signals infinite values
		count_dir_1 = 0;%Number of polled directions (1st order)
		success = 0;
		if decds
			count_dir_2 = 0;
			storvals = zeros(2*n,1);
		else
			storvals = zeros(cardD,1);
		end
%
%%%%%%%%%%%%%%%%%%% First-order loop %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
		while (~success) & (count_dir_1 < cardD)
%
			d = D(:,icycl1);
			xtemp = x + alpha*d;
%
			ftemp = feval(fun,xtemp);
			if isinf(ftemp)
				pbinf=1;
				warning('sounds: Infinite function value - 1st order polling');
			end
%
			storvals(icycl1) = ftemp;
			nevalf = nevalf + 1;
%
			if (ftemp < f-rho1)
				success = 1;
				nits(2) = nits(2) + 1;
			end
%
			count_dir_1 = count_dir_1 + 1;
%
%			Update the cyclicity index in case of failure
			if ~success
				if (icycl1==cardD)
					icycl1 = 1;
				else
					icycl1 = icycl1 + 1;
				end
%
			end
%
		end
%
%%%%%%%%%%%%%%%%%%% Second-order polling(s) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%			Symmetrize the set if needed and poll along the symmetric set
		if ~decds & ~success & ~symmetric
			count_sym = 1;
%
			while ~success & (count_sym <= cardD)
%
				d = D(:,count_sym);
				xtemp = x - alpha*d;
				ftemp = feval(fun,xtemp);
				if isinf(ftemp)
					pbinf = 1;
					warning('sounds: Infinite value - symmetric polling');
				end
				storvals = [storvals;ftemp];
				nevalf = nevalf+1;
%
				if (ftemp < f -rho1)
					success = 1; 
					nits(3) = nits(3) + 1;
				end
				count_sym = count_sym + 1;
			end
		end
%
		if ~strcmp(algo,'sds')
%		
			if decds	
%				Compute directions along the chosen basis
				while (~success) & (count_dir_2 < n)
%
					count_dir_2 = count_dir_2 + 1;
%
					d = B(:,icycl2);
					xplus = x + beta*d;
					xminus = x - beta*d;
%
					fplus = feval(fun,xplus);
					nevalf = nevalf + 1;
					fminus = feval(fun,xminus);	
					nevalf = nevalf + 1;
					if isinf(fplus) || isinf(fminus)
						pbinf = 1;
						warning(strcat('sounds: Infinite value - ',...
						'decoupled symmetric'));
					end
					storvals(count_dir_2) = fplus;
					storvals(count_dir_2+n) = fminus;
%	
					if (fplus < f-rho2) || (fminus < f-rho2)
						success = 2;
						nits(2) = nits(2)+1;
						if (fminus<fplus)
							xtemp = xminus;
							ftemp = fminus;
						else
							xtemp = xplus;
							ftemp = fplus;
						end
					end
%
					if ~success
						if (icycl2==n)
							icycl2 = 1;
						else
							icycl2 = icycl2 + 1;
						end
					end
%
				end
			end
%
%			Choose n directions among all possible if needed and compute 
%			n(n-1)/2 additional points
			cardvals = length(storvals)/2;
			if ~success
				if (typesets(1)==4)
					Ind4 = [p(1:(n-1)) p(n) p(1:(n-1))+cardD p(n)+cardD];
					storvals = storvals(Ind4);
				end
%				Approximation of the Hessian coefficients in the chosen basis
				findiff = storvals(1:cardvals) - 2*f + ...
				storvals((cardvals+1):2*cardvals);
				[orddiff,I] = sort(findiff,'descend');
				orddiff = findiff;
%
%				Set the directions to use among those in B and -B
				minsym = storvals(1:cardvals)-...
				storvals((cardvals+1):(2*cardvals));
				Im = minsym(I(1:n))>0;
				if decds
					Bcross = B(:,I(1:n))*diag(1-Im)-B(:,I(1:n))*diag(Im);
				else
					Dcross = D(:,I(1:n))*diag(1-Im)-D(:,I(1:n))*diag(Im);
				end
				dgvals = storvals(I(1:n)).*(1-Im)+...
				storvals(I(1:n)+cardvals).*Im;
%				Matrix containing the select values for f
				Mevals = diag(dgvals);
%
%				Compute the "crossed" points
				ipt1 = 1;
				ipt2 = 2;
				count_cross = 1;
				total_pts = n*(n-1)/2;
%
				while (~success) & (count_cross <= total_pts)
%
					if decds
						xtemp = x + beta*(Bcross(:,ipt1)+Bcross(:,ipt2));
					else
						xtemp = x + alpha*(Dcross(:,ipt1)+Dcross(:,ipt2));
					end
					ftemp = feval(fun,xtemp);
					if isinf(ftemp)
						warning('sounds: Infinite value - cross polling');
						pbinf = 1;
					end
					nevalf = nevalf+1;
					Mevals(ipt1,ipt2) = ftemp;
%
					if decds & (ftemp < f-rho2)
						success = 2;
						nits(4) = nits(4) + 1;
					elseif ~decds & (ftemp < f-rho1)
						success = 1;
						nits(4) = nits(4) + 1;
					end
%
					if (ipt2<n)
						ipt2 = ipt2 + 1;
					else
						ipt1 = ipt1 + 1;
						ipt2 = 1;
					end
%
					count_cross = count_cross + 1;
				end
%
%				If no successful iterate has been obtained and only finite 
%				values have been obtained, compute an approximate eigenvector
				if decds
					smallstep = (beta < 1e-8);
				else
					smallstep = (alpha < 1e-8);
				end
				if ~success & ~pbinf & ~smallstep
					Devals = zeros(n);
					for i=1:(n-1)
						Devals(i,i+1:n) = Mevals(i,i+1:n)-...
						Mevals(i,i)-dgvals(i+1:n)'+f;
					end
%
					Hevals = Devals+Devals'+diag(orddiff(1:n));
					if decds
						Hevals = 1/(beta^2)*Hevals;
					else
						Hevals = 1/(alpha^2)*Hevals;
					end
					[u,~] = eigs(Hevals,1,'sa');
					u = u/norm(u);
					if decds
						xeplus = x+beta*u;
						xeminus = x-beta*u;
					else
						xeplus = x+alpha*u;
						xeminus = x-alpha*u;
					end
					feplus = feval(fun,xeplus);
					feminus = feval(fun,xeminus);
					nevalf = nevalf + 2;
					if decds &((feplus < f-rho2) || (feminus < f-rho2))
						success = 2;
					elseif ~decds & ((feplus< f-rho1) || (feminus < f-rho1))
						success = 1;
					end
					if success
						nits(5) = nits(5)+1;
						if (feplus<feminus)
							xtemp = xeplus;
							ftemp = feplus;
						else
							xtemp = xeminus;
							ftemp = feminus;
						end
					end
				end
%
			end
%
		end
%
%		Ending the iteration
%	
		if ~success
			alpha = theta*alpha;
			if decds
				beta = thetab*beta;
			end
		else
			f = ftemp;
			x = xtemp;
			hist = [hist [nits(1);x]];
			if (success==1)
				alpha = min(gamma*alpha,alphamax);
			else
				beta = min(gammab*beta,betamax);
			end
%
			if ~isinf(ftarget)
				stopping = stopping || (f > ftarget+tol*diff);
			end
		end
%
%		Stopping boolean update
		stopping = stopping || (alpha < tol_alpha) || ...
		(nevalf >= maxevals-cardD+1);
		if decds
			stopping = stopping || (beta < tol_beta);
		end
%
%		Changing first-order set in the probabilistic case
		if (typesets(1) < 0)
			if typesets(1)==-1
				d = 2*rand(n,1)-1;
				d = d/norm(d);
				D = [d -d];
			else
				for i=1:abs(typesets(1))
					d = 2*rand(n,1)-1;
					D = [D d];
				end
			end
		end
%
	end
%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% End Main Loop
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Final affectations
%
	xsol = x;
	fsol = f;
%
	if isinf(ftarget) || (fsol > ftarget+tol*diff)
		if (alpha >= tol_alpha) 
%			Budget of function evaluations exceeded
			flag = 1;
			if decds
				flag = flag && (beta >= tol_beta);
			end
		else
%			One of the two step sizes became too small
			flag = 2;
		end
	else
%		Desired function accuracy has been reached
		flag = 0;
	end
%
	output.funcCount = nevalf;
	output.nits = nits;
	output.hist = hist;
%			
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Auxiliary functions
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	function D = pss_unifangles(v)
%
%	Computing a Positive Spanning Set of size n+1 with same angles between 
%	each vector.
%
%	------
%	Input:
%	------
%		v: vector of length n; when v is the zero vector, the method performs 
%		a default procedure
%
%	------
%	Output
%	------
%		D: PSS of size n+1 of unitary vectors with uniform angles
%
%	Source:
%	A. R. Conn, K. Scheinberg and L. N. Vicente
%	Introduction to Derivative-Free Optimization
%	MOS-SIAM Series on Optimization
%	SIAM, Philadelphia, 2009.
%
%	Implementation: C. W. Royer
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
		n = length(v);
%		Matrix expressing uniform angles
		A = -1/n*ones(n,n)+(1+1/n)*eye(n);
		C = chol(A);
%
%		Generating random orthogonal matrix
		if norm(v)~=0
			[Q,R] = qr(v);
		else
			Q = eye(n);
		end
%
		D(:,1:n) = Q*C';
		D(:,n+1) = -sum(D(:,1:n),2);
%
	end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	function D = pss_oppbasis(v)
%
%	Computing a Positive Spanning Set of size 2*n in R^n of the form [Q -Q], 
%	where Q is an orthogonal matrix.
%
%	------
%	Input:
%	------
%		v: vector of length n; when v is the zero vector, the method performs 
%		a default procedure
%
%	------
%	Output
%	------
%		D: PSS of size 2*n of the form [Q -Q]
%
%	Source:
%	A. R. Conn, K. Scheinberg and L. N. Vicente
%	Introduction to Derivative-Free Optimization
%	MOS-SIAM Series on Optimization
%	SIAM, Philadelphia, 2009.
%
%	Implementation: C. W. Royer
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%		
		n = length(v);
%		Generating a random rotation matrix
		if norm(v)~=0
			[Q,R] = qr(v);
		else
			Q = eye(n);
			R = 1.1;
		end
%
		if R(1) > 1
			D = Q*[eye(n) -eye(n)];
		else
			D = -Q*[eye(n) -eye(n)];
		end
%
	end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
end
